/* sample gpio device driver for cs-9302 board by k.h. 2010 */

#include <linux/miscdevice.h>
#include <linux/delay.h>
#include <asm/irq.h>

#include <asm/arch/gpio.h>
#include <asm/arch/ep93xx-regs.h>


#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/mm.h>
#include <linux/fs.h>
#include <linux/types.h>
#include <linux/delay.h>
#include <linux/moduleparam.h>
#include <linux/slab.h>
#include <linux/errno.h>
#include <linux/ioctl.h>
#include <linux/cdev.h>
#include <linux/string.h>
#include <linux/list.h>
#include <linux/pci.h>

#include <asm/uaccess.h>
#include <asm/atomic.h>
#include <asm/unistd.h>
#include <asm/gpio.h>
#include <asm/bitops.h>

#define DEVICE_NAME "leds"
static unsigned int dev_major = 0;

extern void gpio_line_config(int line, int direction);
extern void gpio_line_set(int line, int value);
extern void ep93xx_map_io(void);

static unsigned int led_table [] = {
    0,
    1,
    2,
    3,
};

static unsigned int led_cfg_table [] = {
    GPIO_OUT,
    GPIO_OUT,
    GPIO_OUT,
    GPIO_OUT,
};

static int cs_e9302_leds_open(
        struct inode *inode,
        struct file *file )
{
    return 0;
                       }

static int cs_e9302_leds_release(
        struct inode *inode,
        struct file *file )
{
    return 0;
                       }
        

static int cs_e9302_leds_ioctl(
	struct inode *inode, 
	struct file *file, 
	unsigned int cmd, 
	unsigned long arg )
{
    unsigned int data = 0;
    int err = 0;
    int ret = 0;
    unsigned long *addr = 0;
    memset(&data, 0, sizeof(data));

	switch(cmd) {
	case 0:
 	    err = access_ok(VERIFY_READ, (void __user *)arg, sizeof(int));
	    if ( err == 0 ){
		return -EFAULT;
		break;
	    }
	    ret  = copy_from_user(&data, (int __user *)arg, sizeof(data));
	    if  ( ret < 0 ){
		return -EFAULT;
		break;
	    }
	    addr = (unsigned long *)EP93XX_GPIO_REG(0);
	    clear_bit(data, addr);
	    return 0;
	    break;

	case 1:
 	    err = access_ok(VERIFY_READ, (void __user *)arg, sizeof(int));
	    if (err == 0){
		return -EFAULT;
		break;
	    }
	    ret = copy_from_user(&data, (int __user *)arg, sizeof(data));
	    if  ( ret < 0 ){
		return -EFAULT;
		break;
	    }
	    addr = (unsigned long *)EP93XX_GPIO_REG(0);
	    set_bit(data, addr);
	    return 0;
	    break;
	    
	default:
	    return -EINVAL;
	    break;

	}
}

static struct file_operations dev_fops = {
	.owner	=	THIS_MODULE,
        .open   =       cs_e9302_leds_open,
        .release =      cs_e9302_leds_release,
	.ioctl	=	cs_e9302_leds_ioctl,

};


static int __init dev_init(void)
{
        int ret = 0;
	int i;
	int major;
	unsigned long *addr1 = 0;
	unsigned long *addr2 = 0;

	ep93xx_map_io();

        ret = register_chrdev(0, DEVICE_NAME, &dev_fops);
	if ( ret < 0 ){
	    printk("%s driver regist. error\n", DEVICE_NAME );
	    return ret;
	}
	major = ret;
	printk("%s driver(major %d) installed.\n", DEVICE_NAME, major);

	dev_major = major;
	addr1 = (unsigned long *)EP93XX_GPIO_REG(0x10);
	addr2 = (unsigned long *)EP93XX_GPIO_REG(0);
	   
	for (i = 0; i < 4; i++) {
	    set_bit(i, addr1);
	    clear_bit(i, addr2);
	}

	printk (DEVICE_NAME"\tinitialized.\n");

	return ret;
}

static void __exit dev_exit(void)
{
    unregister_chrdev(dev_major, DEVICE_NAME);
}

module_init(dev_init);
module_exit(dev_exit);
MODULE_LICENSE("GPL");
MODULE_AUTHOR("Bonobo Inc.(h.k.)");
